import textwrap
from manim import *
from typing import Sequence

# Define a custom TexTemplate for Devanagari script
_Devanagari = TexTemplate(
    tex_compiler="xelatex",
    output_format=".xdv",
    documentclass="\\documentclass[preview]{standalone}",
    preamble="\\usepackage{fontspec}\n\\usepackage{polyglossia}\n\\usepackage{cancel}\n\\setmainlanguage{english}\n\\setotherlanguage{hindi}\\setmainfont[Script=Devanagari]{Noto Sans}\n\\usepackage{amsmath}\n\\usepackage{amssymb}",
)


def Footer(
    width: float = None,
    stroke_opacity: float = 0,
    fill_color: color = WHITE,
    fill_opacity: float = 1,
    z_index: int = 1,
    height: float = 0.8,
) -> Rectangle:
    """
    Creates a footer rectangle at the bottom of the screen.

    Args:
        width (float, optional): The width of the footer. Defaults to the frame height.
        stroke_opacity (float): The opacity of the stroke. Defaults to 0.
        fill_color (Color): The fill color of the footer. Defaults to WHITE.
        fill_opacity (float): The opacity of the fill. Defaults to 1.
        z_index (int): The z-index of the footer. Defaults to 1.
        height (float): The height of the footer. Defaults to 0.8.

    Returns:
        Rectangle: The footer rectangle object.

    Example:
        footer = Footer(width=10, fill_color=BLUE)
        self.add(footer)
    """
    footer = Rectangle(
        stroke_opacity=stroke_opacity,
        fill_color=fill_color,
        fill_opacity=fill_opacity,
        z_index=z_index,
        width=width if width else config.frame_height,
        height=height,
    )
    footer.to_edge(DOWN, buff=0.0)
    return footer


class Paragraph(Paragraph):
    """
    Custom Paragraph class that wraps text and allows for alignment.

    Args:
        text (Sequence[str]): The text to display as a sequence of strings.
        wrap (bool): Whether to wrap the text. Defaults to True.
        wrap_width (int): The width for wrapping text. Defaults to 50.
        align (str): The alignment of the text. Defaults to "l" (left).
        direction_align (str | None): Optional direction alignment. Defaults to None.
    """

    def __init__(
        self,
        *text: Sequence[str],
        wrap: bool = True,
        wrap_width: int = config.pixel_width,
        align: str = "l",
        **kwargs,
    ):
        self.wrap = wrap
        self.wrap_width = wrap_width
        self.align = kwargs.get("alignment", align)
        self.original_text = map(self.wrap_text, text)
        super().__init__(
            *self.original_text,
            **kwargs,
        )

    def wrap_text(self, text: str) -> str:
        """
        Wraps and justifies the given text based on the specified alignment.

        Args:
            text (str): The text to wrap and justify.

        Returns:
            str: The wrapped and justified text.
        """
        wrapped_text = textwrap.fill(text, width=self.wrap_width)
        # Justify lines based on alignment using match-case
        match self.align:
            case "r" | "right":
                justify_func = str.rjust
            case "c" | "center":
                justify_func = str.center
            case "l" | "left":  # Support for both 'l' and 'left'
                justify_func = str.ljust
            case _:
                justify_func = str.ljust  # Default to left justify
        # Default to left justify
        return "\n".join(
            justify_func(line, self.wrap_width) for line in wrapped_text.splitlines()
        )


class Text(Text):
    """
    Custom Text class that allows for alignment options.

    Args:
        text (str): The text to display.
        align (str): The alignment of the text. Defaults to "l" (left).
        direction_align (str | None): Optional direction alignment. Defaults to None.
    """

    def __init__(
        self,
        text: str,
        align: str = "l",
        **kwargs,
    ):
        self.align = align
        super().__init__(text=text, **kwargs)


class MarkupText(MarkupText):
    """
    Custom MarkupText class that allows for alignment options.

    Args:
        text (str): The text to display.
        align (str): The alignment of the text. Defaults to "l" (left).
        direction_align (str | None): Optional direction alignment. Defaults to None.
    """

    def __init__(
        self,
        text: str,
        align: str = "l",
        **kwargs,
    ):
        self.align = align
        super().__init__(text, **kwargs)


class Tex(Tex):
    """
    Custom Tex class that allows for alignment options.

    Args:
        tex_strings (str): The LaTeX strings to display.
        align (str): The alignment of the text. Defaults to "l" (left).
        direction_align (str | None): Optional direction alignment. Defaults to None.
    """

    def __init__(
        self,
        *tex_strings,
        align: str = "l",
        tex_environment="flushleft",
        **kwargs,
    ):
        self.align = align
        super().__init__(*tex_strings, tex_environment=tex_environment, **kwargs)


class MathTex(MathTex):
    """
    Custom MathTex class that allows for alignment options.

    Args:
        tex_strings (str): The LaTeX strings to display.
        align (str): The alignment of the text. Defaults to "l" (left).
        direction_align (str | None): Optional direction alignment. Defaults to None.
    """

    def __init__(self, *tex_strings, align: str = "l", **kwargs):

        self.align = align
        super().__init__(*tex_strings, **kwargs)


class MathTex_Display(MathTex):
    def __init__(self, *tex_strings, **kwargs):
        super().__init__(*tex_strings, **kwargs)


class Deva_Tex(Tex):
    """Custom Tex class for rendering Devanagari text.

    Args:
        *args: Positional arguments for the Tex class.
        **kwargs: Additional keyword arguments for the Tex class.

    Returns:
        None
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, tex_template=_Devanagari, **kwargs)


class Deva_MathTex(MathTex):
    """Custom MathTex class for rendering Devanagari mathematical expressions.

    Args:
        *args: Positional arguments for the MathTex class.
        **kwargs: Additional keyword arguments for the MathTex class.

    Returns:
        None
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, tex_template=_Devanagari, **kwargs)


class Deva_MathTex_Display(Deva_MathTex):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Deva_Text(Text):
    """Custom Text class for rendering Devanagari text.

    Args:
        *args: Positional arguments for the Text class.
        **kwargs: Additional keyword arguments for the Text class.

    Returns:
        None
    """

    def __init__(self, text: str, font="Noto Sans", **kwargs):
        super().__init__(text, font=font, **kwargs)


class Deva_MarkupText(MarkupText):
    """Custom MarkupText class for rendering Devanagari text with markup support.

    Args:
        *args: Positional arguments for the MarkupText class.
        color (Color): The color of the text. Default is _SET_COLOR.
        font_size (int): The font size for the rendered text. Default is _SET_MARKUPTEXT_FONT_SIZE.
        **kwargs: Additional keyword arguments for the MarkupText class.

    Returns:
        None
    """

    def __init__(self, text: str, font="Noto Sans", **kwargs):
        super().__init__(text, font=font, **kwargs)


class Deva_Paragraph(Paragraph):
    def __init__(self, *text: Sequence[str], font="Noto Sans", **kwargs):
        super().__init__(*text, font=font, **kwargs)


class QuestionText(Text):
    """
    A class representing a question text with language support.

    Args:
        question_no (int): The question number.
        lang (str): The language for the question text. Defaults to "en".
        font_size (int): The font size of the text. Defaults to 25.
        color (Color): The color of the text. Defaults to RED_A.
        weight (str): The weight of the text. Defaults to BOLD.

    Example:
        question = QuestionText(1, lang="en")
        self.add(question)
    """

    def __init__(
        self,
        question_no: int,
        lang: str = "en",
        color=PURE_RED,
        weight=BOLD,
        **kwargs,
    ):
        super().__init__(
            text=self.language(lang, question_no),
            color=color,
            weight=weight,
            **kwargs,
        )

    def language(self, lang, question_no=0):
        """
        Returns the question text in the specified language.

        Args:
            lang (str): The language code.
            question_no (int): The question number.

        Returns:
            str: The formatted question text.
        """
        text = {"hi": "प्रश्‍न"}.get(lang, "Question")
        if question_no:
            text = "{} {} :".format(text, question_no)
        else:
            text = f"{text} :"

        return text


class Deva_QuestionText(QuestionText):
    def __init__(self, question_no: int, lang: str = "hi", **kwargs):
        super().__init__(question_no=question_no, lang=lang, **kwargs)


class AnswerText(Text):
    """
    A class representing an answer text with language support.

    Args:
        question_no (int): The question number.
        lang (str): The language for the answer text. Defaults to "en".
        font_size (int): The font size of the text. Defaults to 25.
        color (Color): The color of the text. Defaults to GREEN_E.
        weight (str): The weight of the text. Defaults to BOLD.

    Example:
        answer = AnswerText(lang="en")
        self.add(answer)
    """

    def __init__(
        self,
        lang="en",
        color=GREEN_E,
        weight=BOLD,
        **kwargs,
    ):
        super().__init__(
            self.language(lang),
            color=color,
            weight=weight,
            **kwargs,
        )

    def language(self, lang):
        """
        Returns the answer text in the specified language.

        Args:
            lang (str): The language code.

        Returns:
            str: The formatted answer text.
        """
        text = {"hi": "उत्तर"}.get(lang, "Answer")
        return f"{text} :"


class Deva_AnswerText(AnswerText):
    def __init__(self, lang: str = "hi", **kwargs):
        super().__init__(lang=lang, **kwargs)


class SolutionText(Text):
    """
    A class representing a solution text with language support.

    Args:
        lang (str): The language for the solution text. Defaults to "en".
        font_size (int): The font size of the text. Defaults to 25.
        color (Color): The color of the text. Defaults to GREEN_E.
        weight (str): The weight of the text. Defaults to BOLD.

    Example:
        solution = SolutionText(lang="en")
        self.add(solution)
    """

    def __init__(
        self,
        lang="en",
        color=GREEN_E,
        weight=BOLD,
        **kwargs,
    ):
        super().__init__(
            text=self.language(lang),
            color=color,
            weight=weight,
            **kwargs,
        )

    def language(self, lang):
        """
        Returns the solution text in the specified language.

        Args:
            lang (str): The language code.

        Returns:
            str: The formatted solution text.
        """
        text = {"hi": "हल"}.get(lang, "Solution")
        return f"{text} :"


class Deva_SolutionText(SolutionText):
    def __init__(self, lang: str = "hi", **kwargs):
        super().__init__(lang=lang, **kwargs)


class Cancel(VGroup):
    """Custom VGroup class for creating a cancel effect on a given mobject.

    Args:
        mobject (Mobject | None): The mobject to be canceled. Default is None.
        stroke_color (Color): The color of the cancel line. Default is RED.
        stroke_width (float): The width of the cancel line. Default is 2.0.
        scale_factor (float): The scale factor for the cancel effect. Default is 1.5.
        **kwargs: Additional keyword arguments for the VGroup class.

    Returns:
        None
    """

    def __init__(
        self,
        mobject: Mobject | None = None,
        stroke_color: ParsableManimColor = RED,
        stroke_width: float = 2.0,
        scale_factor: float = 1.5,
        **kwargs,
    ) -> None:
        super().__init__(Line(UP + RIGHT, DOWN + LEFT), **kwargs)
        if mobject is not None:
            self.replace(mobject, stretch=True)
        self.scale(scale_factor)
        self.set_stroke(color=stroke_color, width=stroke_width)


class BoldTex(Tex):
    def __init__(self, *tex_strings, **kwargs):
        self.tex_string_original = " ".join(tex_strings)
        tex_strings = self.set_bold(*tex_strings)
        super().__init__(*tex_strings, **kwargs)

    def set_bold(self, *tex_strings):
        tex_strings = list(tex_strings)
        tex_strings[0] = r"\textbf{" + tex_strings[0]
        tex_strings[-1] = tex_strings[-1] + r"}"
        return tuple(tex_strings)


class BoldMath(MathTex):
    def __init__(self, *tex_strings, **kwargs):

        self.tex_string_original = " ".join(tex_strings)
        tex_strings = self.set_bold(*tex_strings)
        super().__init__(*tex_strings, **kwargs)

    def set_bold(self, *tex_strings):
        tex_strings = list(tex_strings)
        tex_strings[0] = r"\mathbf{" + tex_strings[0]
        tex_strings[-1] = tex_strings[-1] + r"}"
        return tuple(tex_strings)


class Themes(Scene):
    def set_theme(
        self,
        background_color=BLACK,
        font="sans-serif",
        font_size=DEFAULT_FONT_SIZE,
        set_footer=True,
        footer_color=WHITE,
    ):
        self.Footer = Footer(width=self.camera.frame_width, fill_color=footer_color)
        if set_footer:
            self.add(self.Footer)
        match str(background_color):
            case "#FFFFFF":
                self.camera.background_color = background_color
                _COLOR = BLACK
                Text.set_default(color=_COLOR)
                Paragraph.set_default(color=_COLOR)
                MarkupText.set_default(color=_COLOR)
                Tex.set_default(color=_COLOR)
                MathTex.set_default(color=_COLOR)
                Deva_Tex.set_default(color=_COLOR)
                Deva_MathTex.set_default(color=_COLOR)
                Deva_MarkupText.set_default(color=_COLOR)

        Text.set_default(font=font, font_size=font_size)
        Paragraph.set_default(font=font, font_size=font_size)
        MarkupText.set_default(font=font, font_size=font_size)
        Tex.set_default(font_size=font_size)
        MathTex.set_default(font_size=font_size)
        Deva_Tex.set_default(font_size=font_size)
        Deva_MathTex.set_default(font_size=font_size)
        Deva_MarkupText.set_default(font=font, font_size=font_size)



import re
from manim import *
import manimpango
import functools
from typing import Sequence
from manim_devanagari import (
    Text,
    Paragraph,
    MarkupText,
    Tex,
    MathTex,
    MathTex_Display,
    QuestionText,
    SolutionText,
    AnswerText,
    Deva_Text,
    Deva_MarkupText,
    Deva_Paragraph,
    Deva_Tex,
    Deva_MathTex,
    Deva_MathTex_Display,
    Deva_QuestionText,
    Deva_AnswerText,
    Deva_SolutionText,
)

# @functools.lru_cache(maxsize=None)


def font_list() -> list[str]:
    return manimpango.list_fonts()


def check_font(self, font: str, tuple_value=False) -> bool:
    fonts_list = font_list()
    font_variants = [font, font.lower(), font.capitalize(), font.title()]

    if font.lower() == "sans-serif":
        font = "sans"

    for variant in font_variants:
        if variant in fonts_list:
            return (variant, True) if tuple_value else True

    logger.warning(f"Font {font} not in {fonts_list}.")
    return (font, False) if tuple_value else False


def check_default_font(self, font: str) -> str:
    fonts = [
        font,
        "Lohit Devanagari",
        "Arial Unicode MS",
        "Noto Sans Devanagari",
        "Mangal",
        "Devanagari New",
        "Nirmala UI",
        "Hind",
    ]
    for f in fonts:
        f, check_font = check_font(f, tuple_value=True)
        if check_font:
            return f

    logger.error(f"Font {font} is not font")


def is_hindi(text: str) -> bool:
    """
    Check if the text contains Hindi characters.

    Args:
        text (str): The input text to check.

    Returns:
        bool: True if the text contains Hindi characters, False otherwise.
    """
    return any("\u0900" <= char <= "\u097F" for char in text if char.isalpha())


def is_english(text: str) -> bool:
    """
    Check if the text contains only English characters.

    Args:
        text (str): The input text to check.

    Returns:
        bool: True if the text contains only English alphabetic characters and spaces, False otherwise.
    """
    return all(char.isalpha() or char.isspace() for char in text)


def contains_inline_math(markdown_text):
    # Initialize a counter for dollar signs
    dollar_count = 0
    i = 0

    while i < len(markdown_text):
        # Check for escaped dollar sign
        if markdown_text[i : i + 2] == r"\$":
            i += 2  # Skip the escaped dollar sign
            continue

        # Count dollar signs
        if markdown_text[i] == "$":
            dollar_count += 1

        i += 1

    # Return True if there are an even number of unescaped dollar signs
    return dollar_count % 2 == 0 and dollar_count > 0


def is_latex(text: str) -> bool:
    """
    Check if the text contains LaTeX commands.

    Args:
        text (str): The input text to check.

    Returns:
        bool: True if the text contains LaTeX commands, False otherwise.
    """
    latex_pattern = r"\\[a-zA-Z]+"
    return bool(re.search(latex_pattern, text))


def is_math_mode_inline(text: str) -> bool:
    return text.startswith("$") and text.endswith("$")


def is_math_mode_display(text: str) -> bool:
    return text.startswith("$$") and text.endswith("$$")


def is_html(text):
    # Check if the text contains any HTML-like tags
    if "<" in text and ">" in text:
        # A simple check for well-formedness
        return is_well_formed(text)
    return False


def is_well_formed(html):
    stack = []
    tag = ""
    in_tag = False

    # Remove backslashes for checking
    html = html.replace("\\", "")  # Remove backslashes

    for char in html:
        if char == "<":
            if tag:  # If we were already in a tag, it's malformed
                return False
            in_tag = True
            tag = ""
        elif char == ">":
            if not in_tag:  # If we weren't in a tag, it's malformed
                return False
            in_tag = False
            if tag.startswith("/"):  # Closing tag
                if not stack or stack[-1] != tag[1:]:
                    return False
                stack.pop()
            else:  # Opening tag
                stack.append(tag)
            tag = ""
        elif in_tag:
            tag += char

    return len(stack) == 0  # Stack should be empty if well-formed


def _str_to_mobject_convert(text: str) -> VMobject:
    text_line = config.pixel_width
    if isinstance(text, str):
        is_latex_str = is_latex(text)
        is_str_inline_math = contains_inline_math(text)
        is_hindi_str = is_hindi(text)
        is_math_mode_display_str = is_math_mode_display(text)
        is_markuptext = is_html(text)
        if is_hindi_str:
            if is_math_mode_display_str:
                return Deva_MathTex_Display(text[2:-2].strip())
            elif is_latex_str or is_str_inline_math:
                if is_math_mode_inline(text):
                    return Deva_Tex(text[1:-1].strip())
                return Deva_Tex(text)
            elif is_markuptext:
                return Deva_MarkupText(text)
            else:
                if len(text) <= text_line:
                    return Deva_Text(text)
                return Deva_Paragraph(text)
        else:
            if is_math_mode_display_str:
                return MathTex_Display(text[2:-2].strip())
            elif is_latex_str or is_str_inline_math:
                if is_math_mode_inline(text):
                    return Deva_Tex(text[1:-1].strip())
                return Tex(text)
            elif is_markuptext:
                return MarkupText(text)
            else:
                if len(text) <= text_line:
                    return Text(text)
                return Paragraph(text)

    elif isinstance(text, tuple):
        str_lst = " ".join(text)
        is_latex_str = is_latex(str_lst)
        is_str_inline_math = contains_inline_math(str_lst)
        is_hindi_str = is_hindi(str_lst)
        is_math_mode_display_str = is_math_mode_display(str_lst)
        is_markuptext = is_html(str_lst)
        if is_hindi_str:
            if is_math_mode_display_str:
                return Deva_MathTex_Display(*text[1:-1])
            elif is_latex_str or is_str_inline_math:
                if is_math_mode_inline(str_lst):
                    return Deva_Tex(text[1:-1])
                return Deva_Tex(*text)
            elif is_markuptext:
                return Deva_MarkupText(str_lst)
            else:
                if len(text) > text_line:
                    return Deva_Text(str_lst)
                return Deva_Paragraph(*text)
        else:
            if is_math_mode_display_str:
                return MathTex_Display(*text[1:-1])
            elif is_latex_str or is_str_inline_math:
                if is_math_mode_inline(str_lst):
                    return Deva_Tex(text[1:-1])
                return Tex(*text)
            elif is_markuptext:
                return MarkupText(str_lst)
            else:
                if len(text) > text_line:
                    return Text(str_lst)
                return Paragraph(*text)
    return text


def _str_to_mobject(*vmobjects: VMobject) -> Sequence[VMobject]:
    return tuple(_str_to_mobject_convert(vmobject) for vmobject in vmobjects)


def mobject_to_text(mobject: Mobject) -> str():
    # Tex() or MathTex()
    match type(mobject).__name__:
        case "VGroup":
            m_str = []
            for m in mobject:
                m = mobject_to_text(m)
                if m:
                    m_str.append(m)
            return "\n".join(m_str)
        # Tex
        case "Deva_Tex" | "Tex":
            return mobject.get_tex_string()
        # Text
        case "Deva_Text" | "Text":
            return mobject.original_text
        # Paragraph
        case "Deva_Paragraph" | "Paragraph":
            return mobject.lines_text.original_text
        # MarkupText
        case "Deva_MarkupText" | "MarkupText":
            return mobject.lines_text.original_text
        case _:
            return str(mobject)


def mobject_str(*vmobjects, space=False, **kwargs):
    vmobjects = [mobject_to_text(vmobject) for vmobject in vmobjects]
    return (" " if space else "\n").join(vmobjects)


if __name__ == "__main__":
    # Example usage
    text_samples = [
        "नमस्ते, आप कैसे हैं?",  # Hindi
        "Hello, how are you?",  # English
        r"This is a LaTeX document: \textbf{Bold Text}",  # LaTeX
        "नमस्ते, Hello!",  # Mixed (Hindi and English)
        # Mixed (Hindi, English, LaTeX)
        r"This is a LaTeX document with Hindi: \textbf{नमस्ते}",
    ]

    for sample in text_samples:
        print(f"'{sample}' is of type: {check_text_type(sample)}")



import textwrap
from manim import *
from typing import Iterable, Sequence
from typing_extensions import Self, TypeAlias
from manim.typing import Vector3D
from manim.mobject.opengl.opengl_vectorized_mobject import OpenGLVMobject
from manim_devanagari.helper import (
    _str_to_mobject,
    SolutionText,
    AnswerText,
    QuestionText,
)


"""
This module provides custom Manim classes for creating structured text elements,
footers, and organized groups of questions and answers for educational content.
"""


class Notebook(VMobject):
    def __init__(self, *vmobjects, **kwargs):
        super().__init__(**kwargs)
        vmobjects = _str_to_mobject(*vmobjects)
        self.add(*vmobjects)
        self.arrange_notebook(DOWN, aligned_edge=LEFT)
        self.to_edge(UL)

    def arrange_notebook(
        self,
        direction: Vector3D = RIGHT,
        buff: float = DEFAULT_MOBJECT_TO_MOBJECT_BUFFER,
        center: bool = True,
        **kwargs,
    ) -> Self:
        cue_obj_stack = []
        math_display_mode = None
        for m1, m2 in zip(self.submobjects, self.submobjects[1:]):
            match type(m1).__name__:
                case "cue_column" | "Answer" | "Solution" | "Question":
                    m2.next_to(m1, RIGHT, buff, aligned_edge=UP)
                    cue_obj_stack.append(m1)
                case _:
                    m2.next_to(m1, direction, buff, **kwargs)
                    match type(m2).__name__:
                        case "cue_column" | "Answer" | "Solution" | "Question" if len(
                            cue_obj_stack
                        ):
                            m2.match_x(cue_obj_stack.pop(), RIGHT)

            match type(m1).__name__:
                case (
                    "MathTex"
                    | "MathTex_Display"
                    | "Deva_MathTex"
                    | "Deva_MathTex_Display"
                ):
                    if math_display_mode:
                        m1.match_x(math_display_mode)
                    else:
                        m1.to_edge(ORIGIN + RIGHT - LEFT)
                        math_display_mode = m1

        del cue_obj_stack
        del math_display_mode
        if center:
            self.center()
        return self


class cue_column_Text(Text):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class cue_column_Paragraph(Paragraph):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class cue_column_MarkupText(MarkupText):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class cue_column_Tex(Tex):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class cue_column_MathTex(MathTex):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class cue_column(Text):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Question(QuestionText):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Answer(AnswerText):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)


class Solution(SolutionText):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
